from .attention import Attention
from .attention_layer import AttentionLayer
from .attention_encoder import ResidualAttentionEncoder
from .conditional_attention_encoder import CondtionalResidualAttentionEncoder


import torch
from torch import nn

from neuralDX7.models.attention import Attention



class AttentionLayer(nn.Module):


    def __init__(self, features, hidden_dim, attention):
        """
        features - the number of features the layer has at input and output
        hidden_dim - the hidden dimension of the feedforward network

        """

        super().__init__()

        self.attention = Attention(**attention)
        self.feedforward = nn.Sequential(
            nn.Linear(features, hidden_dim),
            nn.GELU(),
            nn.Linear(hidden_dim, features),
        )

        self.attention_norm = nn.LayerNorm(features)
        self.feedforward_norm = nn.LayerNorm(features, elementwise_affine=False)

    

    def forward(self, X, A):

        X = self.attention_norm(self.attention(X, A) + X)
        X = self.feedforward_norm(self.feedforward(X) + X)

        return X

if __name__=="__main__":

    attention = {
        'n_features': 100,
        'n_hidden': 25,
        'n_heads': 4
    }
    
    model = AttentionLayer(100, 250, attention)

    X = torch.randn(3, 25, 100)
    A = torch.rand(3, 25, 25)>0.5
    Y = model(X, A)

    print(Y.shape)


from torch import nn
from torch.distributions import Normal



class NormalNode(nn.Module):

    def __init__(self, in_features, latent_dim, hidden_dim=None):
        super().__init__()

        if hidden_dim is None:

            hidden_dim = in_features * 2

        self.net = nn.Sequential(
            nn.Linear(in_features, hidden_dim),
            nn.GELU(),
            nn.Linear(hidden_dim, latent_dim * 2)
        )

    def forward(self, x, *args, **kwargs):


        mu, sigma = self.net(x).chunk(2, -1)

        p = Normal(mu, (sigma*0.5).clamp(-5, 4).exp())

        return p






#%%
from collections import namedtuple
from itertools import count
import torch
from torch import nn
from neuralDX7.models.stochastic_nodes import NormalNode

class TriangularSylvester(nn.Module):
    """
    Sylvester normalizing flow with Q=P or Q=I.
    """

    def __init__(self, z_size):

        super(TriangularSylvester, self).__init__()

        self.z_size = z_size
        self.h = nn.Tanh()

        # diag_idx = torch.arange(0, z_size).long()
        # self.register_buffer('diag_idx', diag_idx)

    def der_h(self, x):
        return self.der_tanh(x)

    def der_tanh(self, x):
        return 1 - self.h(x) ** 2

    def forward(self, zk, r1, r2, b, permute_z=None, sum_ldj=True):
        """
        All flow parameters are amortized. conditions on diagonals of R1 and R2 need to be satisfied
        outside of this function.
        Computes the following transformation:
        z' = z + QR1 h( R2Q^T z + b)
        or actually
        z'^T = z^T + h(z^T Q R2^T + b^T)R1^T Q^T
        with Q = P a permutation matrix (equal to identity matrix if permute_z=None)
        :param zk: shape: (batch_size, z_size)
        :param r1: shape: (batch_size, num_ortho_vecs, num_ortho_vecs).
        :param r2: shape: (batch_size, num_ortho_vecs, num_ortho_vecs).
        :param b: shape: (batch_size, 1, self.z_size)
        :return: z, log_det_j
        """
        # Amortized flow parameters
        zk = zk.unsqueeze(1)

        # Save diagonals for log_det_j
        # diag_r1 = r1[:, self.diag_idx, self.diag_idx]
        diag_r1 = torch.diagonal(r1, 0, -1, -2)
        # diag_r2 = r2[:, self.diag_idx, self.diag_idx]
        diag_r2 = torch.diagonal(r2, 0, -1, -2)

        if permute_z is not None:
            # permute order of z
            z_per = zk[:, :, permute_z]
        else:
            z_per = zk

        r2qzb = z_per @ r2.transpose(2, 1) + b
        z = self.h(r2qzb) @ r1.transpose(2, 1)

        if permute_z is not None:
            # permute order of z again back again
            z = z[:, :, permute_z]

        z += zk
        z = z.squeeze(1)

        # Compute log|det J|
        # Output log_det_j in shape (batch_size) instead of (batch_size,1)
        diag_j = diag_r1 * diag_r2
        diag_j = self.der_h(r2qzb).squeeze(1) * diag_j
        diag_j += 1.
        log_diag_j = (diag_j.abs()+1e-8).log()

        if sum_ldj:
            log_det_j = log_diag_j.sum(-1)
        else:
            log_det_j = log_diag_j

        return z, log_det_j

class TriangularSylvesterFlow(nn.Module):
    """
    Variational auto-encoder with triangular Sylvester flows in the encoder. Alternates between setting
    the orthogonal matrix equal to permutation and identity matrix for each flow.
    """

    def __init__(self, in_features, latent_dim, num_flows):

        super().__init__()
        # Initialize log-det-jacobian to zero
        self.log_det_j = 0.

        # Flow parameters
        self.num_flows = num_flows
        self.latent_dim = latent_dim

        # permuting indices corresponding to Q=P (permutation matrix) for every other flow
        flip_idx = torch.arange(latent_dim - 1, -1, -1).long()
        self.register_buffer('flip_idx', flip_idx)

        # self.amor_b = nn.Linear(self.q_z_nn_output_dim, self.num_flows * latent_dim)
        self.q_z = NormalNode(in_features, latent_dim)
        self._flow_params = nn.Linear(in_features,
                self.num_flows * latent_dim * latent_dim + \
                self.num_flows * latent_dim + \
                self.num_flows * latent_dim + \
                self.num_flows * latent_dim
        )
        self.flows = nn.ModuleList([
            TriangularSylvester(latent_dim) for k in range(self.num_flows)
        ])

    def flow_params(self, h):
        """
        Parameterise the base distribution, sample and flow
        """

        batch_size = h.size(0)

        params = self._flow_params(h)
        params = params.reshape(batch_size, self.num_flows, self.latent_dim, -1)
        params = params.transpose(0,1) # batch x flows x z x z  -> flows x batch x z x z
        
        diag1 = torch.tanh(params[...,0])
        diag2 = torch.tanh(params[...,1])
        b = params[...,2].unsqueeze(-2)
        full_d = params[...,3:]

        r1 = torch.triu(full_d, diagonal=1)
        r2 = torch.triu(full_d.transpose(-1, -2), diagonal=1)
        r1 = diag1.diag_embed(0) + r1
        r2 = diag2.diag_embed(0) + r2

        return r1, r2, b

    def forward(self, h, z=None, flow=True):
        """
        Forward pass with orthogonal flows for the transformation z_0 -> z_1 -> ... -> z_k.
        Log determinant is computed as log_det_j = N E_q_z0[\sum_k log |det dz_k/dz_k-1| ].
        """
        Flow = namedtuple('Flow', ('q_z', 'log_det', 'z_0', 'z_k', 'flow'))

        q_z = self.q_z(h)
        z_0 = z_k = q_z.rsample() if z is None else z

        if not flow:
            return Flow(q_z, None, z_0, None)

        r1, r2, b = self.flow_params(h)

        # Sample z_0
        def flow_f(z_k):
            log_det_j = 0.

            # Normalizing flows
            for k, flow_k, r1_k, r2_k, b_k in zip(count(), self.flows, r1, r2, b):

                if k % 2 == 1:
                    # Alternate with reorderering z for triangular flow
                    permute_z = self.flip_idx
                else:
                    permute_z = None

                z_k, log_det_jacobian = flow_k(z_k, r1_k, r2_k, b_k, permute_z, sum_ldj=True)

                log_det_j += log_det_jacobian
            
            return z_k, log_det_j
        z_k, log_det_j = flow_f(z_0)
        return Flow(q_z, log_det_j, z_0, z_k, flow_f)


if __name__=="__main__":
    
    num_ortho_vecs = z_size = 6
    batch_size = 12
    in_features = 64
    
    h = torch.randn(batch_size, in_features)
    # zk = torch.randn(batch_size, z_size)
    # r1 = torch.randn(batch_size, num_ortho_vecs, num_ortho_vecs)
    # r2 = torch.randn(batch_size, num_ortho_vecs, num_ortho_vecs)
    # b = torch.randn(batch_size, 1, z_size)

    f = TriangularSylvesterFlow(in_features, z_size, 3)

    f(h)

# %%



import torch
from torch.nn import functional as F
from importlib import import_module
from torch.optim import AdamW
from torch.distributions.kl import kl_divergence

from agoge import AbstractSolver

from .utils import sigmoidal_annealing

class DX7NeuralProcess(AbstractSolver):

    def __init__(self, model,
        Optim=AdamW, optim_opts=dict(lr= 1e-4),
        max_beta=0.5,
        beta_temp=1e-4,
        **kwargs):

        if isinstance(Optim, str):
            Optim = import_module(Optim)


        self.optim = Optim(params=model.parameters(), **optim_opts)
        self.max_beta = max_beta
        self.model = model

        self.iter = 0
        self.beta_temp = beta_temp

    def loss(self, x, x_hat, x_a, q_context, q_target, z):

        valid_predictions = (~x_a).nonzero().t()

        valid_x_hat = x_hat[(*valid_predictions,)]
        valid_x = x[(*valid_predictions,)]

 
        # kl = kl_divergence(q_target, q_context)#[(*valid_predictions,)]
        # kl = kl_divergence(Normal(torch.zeros_like()), q_context)#[(*valid_predictions,)]
        kl = q_target.log_prob(z) - q_context.log_prob(z)
        kl = kl.sum(-1).mean()
        entropy = q_target.entropy().mean()
        beta = sigmoidal_annealing(self.iter, self.beta_temp).item()

        reconstruction_loss = F.cross_entropy(valid_x_hat, valid_x)
        accuracy = (valid_x_hat.argmax(-1)==valid_x).float().mean()

        loss = reconstruction_loss + 0.25 * beta * kl

        return loss, {
            'accuracy': accuracy,
            'reconstruction_loss': reconstruction_loss,
            'kl': kl,
            'entropy': entropy,
            'beta': beta
        }
        

    def solve(self, x, **kwargs):
        
        x_hat, x_a, q_context, q_target, z  = self.model(x)
        loss, L = self.loss(x, x_hat, x_a, q_context, q_target, z)

        if loss != loss:
            raise ValueError('Nan Values detected')

        if self.model.training:
            self.iter += 1
            self.optim.zero_grad()
            loss.backward()
            self.optim.step()

        return L

    
    def step(self):

        pass


    def state_dict(self):
        
        state_dict = {
            'optim': self.optim.state_dict(),
            'iter': self.iter
        }

        return state_dict

    def load_state_dict(self, state_dict):
        
        self.optim.load_state_dict(state_dict['optim'])
        self.iter = state_dict['iter']


import torch
from torch import nn

from agoge import AbstractModel
from neuralDX7.models.attention import ResidualAttentionEncoder
from neuralDX7.constants import MAX_VALUE, N_PARAMS
from neuralDX7.utils import mask_parameters


class DX7PatchProcess(AbstractModel):

    def __init__(self, features, encoder):
        
        super().__init__()

        self.embedder = nn.Embedding(MAX_VALUE, features)
        self.encoder = ResidualAttentionEncoder(**encoder)

        self.logits = nn.Linear(features, MAX_VALUE)

    def forward(self, X):
        # print(X.shape, )

        # generate random masks
        batch_p = torch.rand(X.shape[0]) # decide p value for each item in batch
        item_logits = torch.rand(X.shape) # random value for each param
        X_a = batch_p.unsqueeze(-1) <= item_logits # active params in X
        X_a = X_a.to(self.device)

        A = (~X_a.unsqueeze(-1)) & (X_a.unsqueeze(-2))
        eye = torch.eye(A.shape[-1]).bool().to(self.device) & (~X_a.unsqueeze(-2))
        A = A | eye
        # 1/0

        X = self.embedder(X) * X_a.unsqueeze(-1).float()
        # X = self.embedder(X) 

        X = self.encoder(X, A)
        # X_hat = mask_parameters(self.logits(X))
        X_hat = self.logits(X)
        # print(X_hat.max(), X_hat.min())

        return X_hat, X_a

    @torch.no_grad()
    def features(self, X):

        X_a = torch.ones_like(X).bool()
        A = X_a.unsqueeze(-1) & X_a.unsqueeze(-2)
        X = self.embedder(X)
        # X = self.embedder(X) 

        X = self.encoder(X, A)

        return X

    @torch.no_grad()
    def generate(self, X, X_a):
        
    
        A = (~X_a.unsqueeze(-1)) & (X_a.unsqueeze(-2))
        eye = torch.eye(A.shape[-1]).bool().to(self.device) & (~X_a.unsqueeze(-2))
        A = A | eye

        X = self.embedder(X) * X_a.unsqueeze(-1).float()
        X = self.encoder(X, A)
        X_hat = mask_parameters(self.logits(X))

        X_hat = torch.distributions.Categorical(logits=X_hat)

        return X_hat



if __name__=='__main__':

    layer_features = 100
    n_heads = 4
    N_PARAMS = 8

    head_features = layer_features // n_heads

    attention = {
        'n_features': layer_features,
        'n_hidden': head_features,
        'n_heads': n_heads
    }
    
    attention_layer = {
        'attention': attention,
        'features': layer_features,
        'hidden_dim': 555
    }

    encoder = {
        'features': layer_features,
        'attention_layer': attention_layer,
        'max_len': N_PARAMS
    }
        
    
    model = DX7PatchProcess(layer_features, encoder=encoder)
    X = torch.distributions.Categorical(torch.ones(128)).sample((3, N_PARAMS))

    logits = model(X)
    print(logits.shape)
    print(logits[0])






import torch
from torch import nn
from torch.nn import functional as F

from agoge import AbstractModel
from neuralDX7.models.attention import ResidualAttentionEncoder, CondtionalResidualAttentionEncoder
from neuralDX7.models.general import FeedForwardGELU
from neuralDX7.models.stochastic_nodes import NormalNode
from neuralDX7.constants import MAX_VALUE, N_PARAMS
from neuralDX7.utils import mask_parameters


class DX7NeuralProcess(AbstractModel):

    def __init__(self, features, latent_dim, encoder, decoder, deterministic_path_drop_rate=0.5):
        
        super().__init__()

        self.embedder = nn.Embedding(MAX_VALUE, features)
        self.encoder = ResidualAttentionEncoder(**encoder)
        self._latent_encoder = nn.ModuleList([
            ResidualAttentionEncoder(**encoder),
            NormalNode(features, latent_dim)]
        )
        self.z_to_c = nn.Linear(latent_dim, latent_dim*155)
        self.decoder = CondtionalResidualAttentionEncoder(**decoder)
        self.logits = FeedForwardGELU(features, MAX_VALUE)
        self.drop = nn.Dropout(deterministic_path_drop_rate)

    def latent_encoder(self,  X, A, mean=False):

        encoder, q_x = self._latent_encoder

        return q_x(encoder(X, A).mean(-2))


    def forward(self, X):

        # generate random masks
        batch_p = torch.rand(X.shape[0]) # decide p value for each item in batch
        item_logits = torch.rand(X.shape) # random value for each param
        X_a = batch_p.unsqueeze(-1) <= item_logits # active params in X
        X_a = X_a.to(self.device)

        A = (~X_a.unsqueeze(-1)) & (X_a.unsqueeze(-2))
        eye = torch.eye(A.shape[-1]).bool().to(self.device) & (~X_a.unsqueeze(-2))
        A = A | eye
        # A = A | True

        X_target = self.embedder(X)
        
        X_context = X_target * X_a.unsqueeze(-1).float()

        q_context = self.latent_encoder(X_context, A)
        q_target = self.latent_encoder(X_target, A | (~X_a.unsqueeze(-1)))

        # r = self.drop(self.encoder(X_context, A))
        # X_encoded = F.drop out
        z = q_target.rsample()
        # z_context = q_context.rsample()
        # mask = (torch.rand_like(z_target[...,[0]]) > 0.5).float()
        # z = (z_target * mask) + (z_context * (1-mask))

        c = self.z_to_c(z).view(z.shape[0], 155, -1)
        # c = z.unsqueeze(-2)

        X_dec = self.decoder(X_context, A, c)
        X_hat = self.logits(X_dec)

        return X_hat, X_a, q_context, q_target, z

    @torch.no_grad()
    def features(self, X, X_a):

        A = (~X_a.unsqueeze(-1)) & (X_a.unsqueeze(-2))
        eye = torch.eye(A.shape[-1]).bool().to(self.device) & (~X_a.unsqueeze(-2))
        A = A | eye

        X = self.embedder(X) * X_a.unsqueeze(-1).float()
        q = self.latent_encoder(X, A)

        return q

    @torch.no_grad()
    def generate_z(self, X, X_a, z, t=1.):


        A = (~X_a.unsqueeze(-1)) & (X_a.unsqueeze(-2))
        eye = torch.eye(A.shape[-1]).bool().to(self.device) & (~X_a.unsqueeze(-2))
        A = A | eye

        X = self.embedder(X)
        X = X * X_a.unsqueeze(-1).float()       

        c = self.z_to_c(z).view(z.shape[0], 155, -1)
        X_dec = self.decoder(X, A, c)
        X_hat = mask_parameters(self.logits(X_dec))
        X_hat = torch.distributions.Categorical(logits=X_hat/t)

        return X_hat

    @torch.no_grad()
    def generate(self, X, X_a, sample=True, t=1.):

        q = self.features(X, X_a)

        

        z = q.sample()

        c_gamma, c_beta = self.z_to_c(z).chunk(2, -1)


        X_hat = mask_parameters(self.logits(c_gamma))

        X_hat = torch.distributions.Categorical(logits=X_hat/t)

        return X_hat



if __name__=='__main__':

    layer_features = 100
    n_heads = 4

    head_features = layer_features // n_heads

    attention = {
        'n_features': layer_features,
        'n_hidden': head_features,
        'n_heads': n_heads
    }
    
    attention_layer = {
        'attention': attention,
        'features': layer_features,
        'hidden_dim': 555
    }

    encoder = {
        'features': layer_features,
        'attention_layer': attention_layer,
        'max_len': N_PARAMS
    }
        
    
    model = DX7PatchProcess(layer_features, encoder=encoder)
    X = torch.distributions.Categorical(torch.ones(128)).sample((3, N_PARAMS))

    logits = model(X)
    print(logits.shape)
    print(logits[0])



import torch
from torch import nn

class Attention(nn.Module):



    def __init__(self, n_features, n_hidden, n_heads=8, inf=1e9):
        """
        n_features - number of input features
        n_hidden - hidden dim per head
        n_heads - number of heads
        """

        super().__init__()

        self.QKV = nn.Linear(n_features, n_hidden * 3 * n_heads)
        self.n_heads = n_heads
        self._inf = inf

    @property
    def inf(self):
        # if self.training:
            return self._inf
        # return float('inf')

    def forward(self, X, A):

        *input_shape, _ = X.shape


        QKV = self.QKV(X).reshape(*input_shape, -1, 3, self.n_heads)
        n_dims = len(QKV.shape)
        permuter = torch.arange(n_dims).roll(2)
        Q, K, V = QKV.permute(*permuter)

        qk_t = (Q @ K.transpose(-1, -2)) / (self.n_heads**(1/2))
        qk_t_masked =  qk_t.masked_fill(~A, -self.inf)
        
        Y = qk_t_masked.softmax(-1) @ V

        n_dims = len(Y.shape)
        permuter = torch.arange(n_dims).roll(-1)
        Y = Y.permute(*permuter).flatten(-2, -1)
        return Y



if __name__=="__main__":



    model = Attention(100, 20)
    X = torch.randn(3, 25, 100)
    A = torch.rand(3, 25, 25)>0.5
    Y = model(X, A)
    print(Y.shape)



import torch
from torch import nn
from torch.nn import functional as F

from agoge import AbstractModel
from neuralDX7.models.attention import ResidualAttentionEncoder, CondtionalResidualAttentionEncoder
from neuralDX7.models.general import FeedForwardGELU
from neuralDX7.models.stochastic_nodes import TriangularSylvesterFlow
from neuralDX7.constants import MAX_VALUE, N_PARAMS
from neuralDX7.utils import mask_parameters


class DX7VAE(AbstractModel):

    def __init__(self, features, latent_dim, encoder, decoder, deterministic_path_drop_rate=0.5,  num_flows=3):
        
        super().__init__()

        self.embedder = nn.Embedding(MAX_VALUE, features)
        self.encoder = ResidualAttentionEncoder(**encoder)
        self._latent_encoder = nn.ModuleList([
            ResidualAttentionEncoder(**encoder),
            TriangularSylvesterFlow(features, latent_dim, num_flows)]
        )
        self.z_to_c = nn.Linear(latent_dim, latent_dim*155)
        self.decoder = CondtionalResidualAttentionEncoder(**decoder)
        self.logits = FeedForwardGELU(features, MAX_VALUE)


        self.drop = nn.Dropout(deterministic_path_drop_rate)
        self.n_features = features

    def latent_encoder(self,  X, A, z=None, mean=False):

        encoder, q_x = self._latent_encoder

        return q_x(encoder(X, A).mean(-2), z)


    def forward(self, X):

        batch_size = X.shape[0]

        A = torch.ones_like(X).bool()
        A = A[...,None] | A[...,None,:]

        X_emb = self.embedder(X)
        
        flow = self.latent_encoder(X_emb, A)
        
        c = self.z_to_c(flow.z_k).view(batch_size, 155, -1)
        # c = z.unsqueeze(-2)

        X_dec = self.decoder(torch.ones_like(X_emb), A, c)
        X_hat = self.logits(X_dec)

        return {
            'X_hat': X_hat,
            'flow': flow,
        }


    @torch.no_grad()
    def features(self, X):
        # pass
        A = torch.ones_like(X).bool()
        A = A[...,None] | A[...,None,:]
        # eye = torch.eye(A.shape[-1]).bool().to(self.device) & (~X_a.unsqueeze(-2))
        # A = A | eye

        X = self.embedder(X)
        q = self.latent_encoder(X, A)

        return q.q_z

    @torch.no_grad()
    def generate(self, z, t=1.):
        
        
        A = z.new(z.size(0), 155, 155).bool() | 1
        X = z.new(z.size(0), 155, self.n_features)
        X = X * 0 + 1

        c = self.z_to_c(z).view(z.shape[0], 155, -1)
        X_dec = self.decoder(X, A, c)
        X_hat = mask_parameters(self.logits(X_dec))
        X_hat = torch.distributions.Categorical(logits=X_hat/t)

        return X_hat

    # @torch.no_grad()
    # def generate(self, X, X_a, sample=True, t=1.):

    #     q = self.features(X, X_a)

        

    #     z = q.sample()

    #     c_gamma, c_beta = self.z_to_c(z).chunk(2, -1)


    #     X_hat = mask_parameters(self.logits(c_gamma))

    #     X_hat = torch.distributions.Categorical(logits=X_hat/t)

    #     return X_hat



if __name__=='__main__':

    layer_features = 100
    n_heads = 4

    head_features = layer_features // n_heads

    attention = {
        'n_features': layer_features,
        'n_hidden': head_features,
        'n_heads': n_heads
    }
    
    attention_layer = {
        'attention': attention,
        'features': layer_features,
        'hidden_dim': 555
    }

    encoder = {
        'features': layer_features,
        'attention_layer': attention_layer,
        'max_len': N_PARAMS
    }
        
    
    model = DX7PatchProcess(layer_features, encoder=encoder)
    X = torch.distributions.Categorical(torch.ones(128)).sample((3, N_PARAMS))

    logits = model(X)
    print(logits.shape)
    print(logits[0])




import torch
import numpy as np

def position_encoding_init(n_position, emb_dim):
    ''' Init the sinusoid position encoding table '''

    # keep dim 0 for padding token position encoding zero vector
    position_enc = np.array([
        [pos / np.power(10000, 2 * (j // 2) / emb_dim) for j in range(emb_dim)]
        if pos != 0 else np.zeros(emb_dim) for pos in range(n_position)])
    

    position_enc[1:, 0::2] = np.sin(position_enc[1:, 0::2]) # apply sin on 0th,2nd,4th...emb_dim
    position_enc[1:, 1::2] = np.cos(position_enc[1:, 1::2]) # apply cos on 1st,3rd,5th...emb_dim
    return torch.from_numpy(position_enc).type(torch.FloatTensor)



import torch
from torch.nn import functional as F
from importlib import import_module
from torch.optim import AdamW
from torch.distributions.kl import kl_divergence
from torch.distributions import Normal
from agoge import AbstractSolver

from .utils import sigmoidal_annealing

class DX7VAE(AbstractSolver):
 
    def __init__(self, model,
        Optim=AdamW, optim_opts=dict(lr= 1e-4),
        max_beta=0.5,
        beta_temp=1e-4,
        **kwargs):

        if isinstance(Optim, str):
            Optim = import_module(Optim)

        self.optim = Optim(params=model.parameters(), **optim_opts)
        self.max_beta = max_beta
        self.model = model

        self.iter = 0
        self.beta_temp = beta_temp

    def loss(self, X, X_hat, flow):

    
        p_z_k = Normal(0,1).log_prob(flow.z_k).sum(-1)
        q_z_0 = flow.q_z.log_prob(flow.z_0).sum(-1)
        kl = (q_z_0-p_z_k-flow.log_det).mean() / flow.z_k.shape[-1]

        beta = sigmoidal_annealing(self.iter, self.beta_temp).item()

        reconstruction_loss = F.cross_entropy(X_hat.transpose(-1, -2), X)
        accuracy = (X_hat.argmax(-1)==X).float().mean()

        loss = reconstruction_loss + self.max_beta * beta * kl

        return loss, {
            'accuracy': accuracy,
            'reconstruction_loss': reconstruction_loss,
            'kl': kl,
            'beta': beta,
            'log_det': flow.log_det.mean(),
            'p_z_k': p_z_k.mean(),
            'q_z_0': q_z_0.mean(),
            # 'iter': self.iter // self.
        }
        

    def solve(self, X, **kwargs):
        
        Y = self.model(**X)
        loss, L = self.loss(**X, **Y)

        if loss != loss:
            raise ValueError('Nan Values detected')

        if self.model.training:
            self.iter += 1
            self.optim.zero_grad()
            loss.backward()
            self.optim.step()

        return L

    
    def step(self):

        pass


    def state_dict(self):
        
        state_dict = {
            'optim': self.optim.state_dict(),
            'iter': self.iter
        }

        return state_dict

    def load_state_dict(self, state_dict):
        
        self.optim.load_state_dict(state_dict['optim'])
        self.iter = state_dict['iter']


import torch
from torch.nn import functional as F
from importlib import import_module
from torch.optim import AdamW

from agoge import AbstractSolver



class DX7PatchProcess(AbstractSolver):

    def __init__(self, model,
        Optim=AdamW, optim_opts=dict(lr= 1e-4),
        max_beta=0.5,
        **kwargs):

        if isinstance(Optim, str):
            Optim = import_module(Optim)


        self.optim = Optim(params=model.parameters(), **optim_opts)
        self.max_beta = max_beta
        self.model = model

    def loss(self, x, x_hat, x_a):

        valid_predictions = (~x_a).nonzero().t()

        valid_x_hat = x_hat[(*valid_predictions,)]
        valid_x = x[(*valid_predictions,)]

        reconstruction_loss = F.cross_entropy(valid_x_hat, valid_x)
        accuracy = (valid_x_hat.argmax(-1)==valid_x).float().mean()

        return reconstruction_loss, {
            'accuracy': accuracy,
            'reconstruction_loss': reconstruction_loss,
        }
        

    def solve(self, x, **kwargs):
        
        x_hat, x_a  = self.model(x)
        loss, L = self.loss(x, x_hat, x_a)

        if loss != loss:
            raise ValueError('Nan Values detected')

        if self.model.training:

            self.optim.zero_grad()
            loss.backward()
            self.optim.step()
        
        return L

    
    def step(self):

        pass


    def state_dict(self):
        
        state_dict = {
            'optim': self.optim.state_dict()
        }

        return state_dict

    def load_state_dict(self, state_dict):
        
        self.optim.load_state_dict(state_dict['optim'])


from pathlib import Path
import numpy as np
import torch
from neuralDX7 import DEFAULTS




class DX7SysexDataset():
    

    def __init__(self, data_file='dx7.npy', root=DEFAULTS['ARTIFACTS_ROOT'], data_size=1.):

        assert data_size <= 1

        self.data_size = data_size

        if not isinstance(root, Path):
            root = Path(root).expanduser()

        self.data = np.load(root.joinpath(data_file)) 

    def __getitem__(self, index):

        item = torch.tensor(self.data[index].item()).long()

        return {'X': item}
    
    def __len__(self):
        return int(len(self.data) * self.data_size)


if __name__ == "__main__":
    

    dataset = DX7SysexDataset()

    print([dataset[i] for i in np.random.randint(0, len(dataset)-1, 20)])


#!/usr/bin/env python
from setuptools import setup, find_packages
from pathlib import Path

root_path = Path(__file__).parent


readme = Path('README.md').read_text()
requirements = Path('requirements.txt').read_text().split('\n')
__version__ = Path('version').read_text().strip()

setup(
    name='neuralDX7',
    packages = find_packages(),
    license='MIT',
    version=__version__,
    author='Nintorac',
    author_email='neuralDX7@nintorac.dev',
    url='https://github.com/nintorac/neuralDX7',
    description='Models related to the DX7 fm synthesizer',
    long_description=readme,
    zip_safe=True,
    install_requires=requirements,
    classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
  long_description_content_type='text/markdown'
)


import torch


def sigmoidal_annealing(iter_nb, t=1e-4, s=-6):
    """

    iter_nb - number of parameter updates completed
    t - step size
    s - slope of the sigmoid
    """
    
    t, s = torch.tensor(t), torch.tensor(s).float()
    x0 = torch.sigmoid(s)
    value = (torch.sigmoid(iter_nb*t + s) - x0)/(1-x0) 

    return value


from pathlib import Path
import bitstruct
import mido


def take(take_from, n):
    for _ in range(n):
        yield next(take_from)

N_OSC = 6
N_VOICES = 32

def checksum(data):
    return (128-sum(data)&127)%128

GLOBAL_VALID_RANGES = {
    'PR1':  range(0, 99+1),
    'PR2':  range(0, 99+1),
    'PR3':  range(0, 99+1),
    'PR4':  range(0, 99+1),
    'PL1':  range(0, 99+1),
    'PL2':  range(0, 99+1),
    'PL3':  range(0, 99+1),
    'PL4':  range(0, 99+1),
    'ALG':  range(0, 31+1),
    'OKS':  range(0, 1+1),
    'FB':   range(0, 7+1),
    'LFS':  range(0, 99+1),
    'LFD':  range(0, 99+1),
    'LPMD':  range(0, 99+1),
    'LAMD':  range(0, 99+1),
    'LPMS': range(0, 7+1),
    'LFW':  range(0, 5+1),
    'LKS':  range(0, 1+1),
    'TRNSP':  range(0, 48+1),
    'NAME CHAR 1': range(128),
    'NAME CHAR 2': range(128),
    'NAME CHAR 3': range(128),
    'NAME CHAR 4': range(128),
    'NAME CHAR 5': range(128),
    'NAME CHAR 6': range(128),
    'NAME CHAR 7': range(128),
    'NAME CHAR 8': range(128),
    'NAME CHAR 9': range(128),
    'NAME CHAR 10': range(128),
 }

OSCILLATOR_VALID_RANGES = {
    'R1':  range(0, 99+1),
    'R2':  range(0, 99+1),
    'R3':  range(0, 99+1),
    'R4':  range(0, 99+1),
    'L1':  range(0, 99+1),
    'L2':  range(0, 99+1),
    'L3':  range(0, 99+1),
    'L4':  range(0, 99+1),
    'BP':  range(0, 99+1),
    'LD':  range(0, 99+1),
    'RD':  range(0, 99+1),
    'RC':  range(0, 3+1),
    'LC':  range(0, 3+1),
    'DET': range(0, 14+1),
    'RS':  range(0, 7+1),
    'KVS': range(0, 7+1),
    'AMS': range(0, 3+1),
    'OL':  range(0, 99+1),
    'FC':  range(0, 31+1),
    'M':   range(0, 1+1),
    'FF':  range(0, 99+1),
}

VOICE_PARAMETER_RANGES = {f'{i}_{key}': value for key, value in OSCILLATOR_VALID_RANGES.items() for i in range(N_OSC)}
VOICE_PARAMETER_RANGES.update(GLOBAL_VALID_RANGES)

def verify(actual, ranges):
    assert set(actual.keys())==set(ranges.keys()), 'Params dont match'
    for key in actual:
        if not actual[key] in ranges[key]:
            return False
    return True


HEADER_KEYS = [
    'ID',
    'Sub-status',
    'format number',
    'byte count',
    'byte count',
]

GENERAL_KEYS = [
    'PR1',
    'PR2',
    'PR3',
    'PR4',
    'PL1',
    'PL2',
    'PL3',
    'PL4',
    'ALG',
    'OKS',
    'FB',
    'LFS',
    'LFD',
    'LPMD',
    'LAMD',
    'LPMS',
    'LFW',
    'LKS',
    'TRNSP',
    'NAME CHAR 1',
    'NAME CHAR 2',
    'NAME CHAR 3',
    'NAME CHAR 4',
    'NAME CHAR 5',
    'NAME CHAR 6',
    'NAME CHAR 7',
    'NAME CHAR 8',
    'NAME CHAR 9',
    'NAME CHAR 10',
]

OSC_KEYS = [
    'R1',
    'R2',
    'R3',
    'R4',
    'L1',
    'L2',
    'L3',
    'L4',
    'BP',
    'LD',
    'RD',
    'RC',
    'LC',
    'DET',
    'RS',
    'KVS',
    'AMS',
    'OL',
    'FC',
    'M',
    'FF',
]

FOOTER_KEYS = ['checksum']


VOICE_KEYS = [f'{i}_{key}' for i in range(6) for key in OSC_KEYS] + \
        GENERAL_KEYS 

KEYS =  HEADER_KEYS + \
        list(VOICE_KEYS * N_VOICES) + \
        FOOTER_KEYS



header_bytes = [
    'p1u7',             # ID # (i=67; Yamaha)
    'p1u7',             # Sub-status (s=0) & channel number (n=0; ch 1)
    'p1u7',             # format number (f=9; 32 voices)
    'p1u7',             # byte count MS byte
    'p1u7',             # byte count LS byte (b=4096; 32 voices)
]




general_parameter_bytes = [ 
    'p1u7',             # PR1
    'p1u7',             # PR2
    'p1u7',             # PR3
    'p1u7',             # PR4
    'p1u7',             # PL1
    'p1u7',             # PL2
    'p1u7',             # PL3
    'p1u7',             # PL4
    'p3u5',             # ALG
    'p4u1u3',           # OKS|    FB
    'p1u7',             # LFS
    'p1u7',             # LFD
    'p1u7',             # LPMD
    'p1u7',             # LAMD
    'p1u3u3u1',         # LPMS |      LFW      |LKS
    'p1u7',             # TRNSP
    'p1u7',             # NAME CHAR 1
    'p1u7',             # NAME CHAR 2
    'p1u7',             # NAME CHAR 3
    'p1u7',             # NAME CHAR 4
    'p1u7',             # NAME CHAR 5
    'p1u7',             # NAME CHAR 6
    'p1u7',             # NAME CHAR 7
    'p1u7',             # NAME CHAR 8
    'p1u7',             # NAME CHAR 9
    'p1u7',             # NAME CHAR 10
]

osc_parameter_bytes = [
    'p1u7',         # R1
    'p1u7',         # R2
    'p1u7',         # R3
    'p1u7',         # R4
    'p1u7',         # L1
    'p1u7',         # L2
    'p1u7',         # L3
    'p1u7',         # L4
    'p1u7',         # BP
    'p1u7',         # LD
    'p1u7',         # RD
    'p4u2u2',       # RC | LC 
    'p1u4u3',       # DET | RS
    'p3u3u2',       # KVS | AMS
    'p1u7',         # OL
    'p2u5u1',       # FC | M
    'p1u7'          # FF
]

voice_bytes = (osc_parameter_bytes * N_OSC) + general_parameter_bytes

tail_bytes = [
    'p1u7',         # checksum
]


full_string = ''.join(header_bytes + osc_parameter_bytes * 6 + general_parameter_bytes)
dx7_struct = bitstruct.compile(full_string)

voice_struct = bitstruct.compile(''.join(voice_bytes), names=VOICE_KEYS)
header_struct = bitstruct.compile(''.join(header_bytes))

N_PARAMS = len(VOICE_PARAMETER_RANGES)
MAX_VALUE = max([max(i) for i in VOICE_PARAMETER_RANGES.values()]) + 1


"""
SYSEX Message: Bulk Data for 1 Voice
------------------------------------
       bits    hex  description

     11110000  F0   Status byte - start sysex
     0iiiiiii  43   ID # (i=67; Yamaha)
     0sssnnnn  00   Sub-status (s=0) & channel number (n=0; ch 1)
     0fffffff  00   format number (f=0; 1 voice)
     0bbbbbbb  01   byte count MS byte
     0bbbbbbb  1B   byte count LS byte (b=155; 1 voice)
     0ddddddd  **   data byte 1

        |       |       |

     0ddddddd  **   data byte 155
     0eeeeeee  **   checksum (masked 2's complement of sum of 155 bytes)
     11110111  F7   Status - end sysex



///////////////////////////////////////////////////////////
"""
class DX7Single():
    HEADER = int('0x43', 0), int('0x00', 0), int('0x00', 0), int('0x01', 0), int('0x1B', 0)
    
    GENERAL_KEYS = [
        'PR1',
        'PR2',
        'PR3',
        'PR4',
        'PL1',
        'PL2',
        'PL3',
        'PL4',
        'ALG',
        'FB',
        'OKS',
        'LFS',
        'LFD',
        'LPMD',
        'LAMD',
        'LKS',
        'LFW',
        'LPMS',
        'TRNSP',
        'NAME CHAR 1',
        'NAME CHAR 2',
        'NAME CHAR 3',
        'NAME CHAR 4',
        'NAME CHAR 5',
        'NAME CHAR 6',
        'NAME CHAR 7',
        'NAME CHAR 8',
        'NAME CHAR 9',
        'NAME CHAR 10',
    ]

    OSC_KEYS = [
        'R1',
        'R2',
        'R3',
        'R4',
        'L1',
        'L2',
        'L3',
        'L4',
        'BP',
        'LD',
        'RD',
        'LC',
        'RC',
        'RS',
        'AMS',
        'KVS',
        'OL',
        'M',
        'FC',
        'FF',
        'DET',
    ]

    @staticmethod
    def keys():

        osc_keys = DX7Single.OSC_KEYS
        osc_params = [f'{i}_{param}' for i in range(N_OSC) for param in osc_keys]
        # print(osc_params)
        all = osc_params + DX7Single.GENERAL_KEYS
        return all

    @staticmethod
    def struct():
        return bitstruct.compile('p1u7'*155, names=DX7Single.keys())


    @staticmethod
    def to_syx(voices):


        assert len(voices)==1
        voice = voices[0]
        voices_bytes = bytes()
        voices_bytes = DX7Single.struct().pack(dict(zip(VOICE_KEYS, voice)))    
        
        patch_checksum = [checksum(voices_bytes)]

        data = bytes(DX7Single.HEADER) \
            + voices_bytes \
            + bytes(patch_checksum)


        return mido.Message('sysex', data=data)
        

def consume_syx(path):

    path = Path(path).expanduser()
    try:
        preset = mido.read_syx_file(path.as_posix())[0]
    except IndexError as e:
        return None
    except ValueError as e:
        return None
    if len(preset.data) == 0:
        return None

    def get_voice(data):
        
        unpacked = voice_struct.unpack(data)

        if not verify(unpacked, VOICE_PARAMETER_RANGES):
            return None
        
        return unpacked

    get_header = header_struct.unpack
    sysex_iter = iter(preset.data)
    
    try:
        header = get_header(bytes(take(sysex_iter, len(header_bytes))))
        yield from (get_voice(bytes(take(sysex_iter, len(voice_bytes)))) for _ in range(N_VOICES))
    except RuntimeError:
        return None

if __name__=="__main__":
    print(VOICE_KEYS)

    # print(DX7Single.to_syx(n)


import torch
from torch import nn

class FeedForwardGELU(nn.Module):


    def __init__(self, features, out_features=None, exapnsion_factor=3):

        super().__init__()
        out_features = features if out_features is None else out_features

        # self.net = nn.ModuleList(
        self.net = nn.Sequential(
            # [
            nn.Linear(features, features*exapnsion_factor),
            nn.GELU(),
            nn.Linear(features*exapnsion_factor, out_features)
            # ]
        )

    def forward(self, x):

        return self.net(x)


import torch
from os import environ

from torch import nn

from agoge import AbstractModel
from neuralDX7.models.attention import AttentionLayer
from neuralDX7.models.general import FeedForwardGELU
from neuralDX7.models.utils import position_encoding_init



class CondtionalResidualAttentionEncoder(AbstractModel):
    def __init__(self, features, c_features, attention_layer, max_len=200, n_layers=3):
        super().__init__()


        self.layers = nn.ModuleList(
            map(lambda x: AttentionLayer(**attention_layer), range(n_layers))
        )

        positional_encoding = position_encoding_init(max_len, features)
        self.c_layers = nn.ModuleList(
            map(lambda x: FeedForwardGELU(c_features, features*2), range(n_layers))
        )

        self.p2x = nn.Linear(features, features * 2)
        self.register_buffer('positional_encoding', positional_encoding)


    def forward(self, X, A, c):
        encodings = self.p2x(self.positional_encoding).chunk(2, -1)
        
        fs = f_gamma, f_beta = torch.sigmoid, torch.tanh
        gamma_p, beta_p = map(lambda f, x: f(x), fs, encodings)
        # gamma = torch.sigmoid(gamma)
        # beta = torch.tanh(beta)

        X = gamma_p * X + beta_p

        for layer, c_layer in zip(self.layers, self.c_layers):

            conditioning = c_layer(c).chunk(2, -1)
            gamma_c, beta_c = map(lambda f, x: f(x), fs, conditioning)

            X = layer(gamma_c * X + beta_c, A)

        return X
        


if __name__=='__main__':

    layer_features = 100
    n_heads = 4

    head_features = layer_features // n_heads

    attention = {
        'n_features': layer_features,
        'n_hidden': head_features,
        'n_heads': n_heads
    }
    
    attention_layer = {
        'attention': attention,
        'features': layer_features,
        'hidden_dim': 555
    }

    
    max_len = 25
    
    
    model = ResidualAttentionEncoder(layer_features, attention_layer, max_len=max_len)
    A = torch.rand(3, 25, 25)>0.5
    X = torch.distributions.Categorical(torch.ones(128)).sample((3, 25))

    model(X, A)

